# wordnet.py, version 1.2
#
# Author: Oliver Steele, steele@cs.brandeis.edu
# Source: http://www.cs.brandeis.edu/~steele/sources/wordnet.py
# Home Page: http://www.cs.brandeis.edu/~steele/sources/wordnet-python.html
#
# Copyright (c) 1998-1999 by Oliver Steele.  Use this however you like,
# so long as you preserve the attribution and label your changes.

"""Module wordnet -- an interface to the Wordnet 1.6 database.

INSTALLATION

- Put this file where Python can see it.
- Download the WordNet database files from http://www.cogsci.princeton.edu/~wn/.
- Either:
    - Set the WNHOME environment variable to the pathname of WordNet's "Dictionary" or "dict"
      directory.
  or:
	- Set the WNSEARCHDIR environment variable to the pathname of the directory that _contains_
	  this directory.
  or:
    - Place an alias to WordNet's "Dictionary" (Mac) or "dict" (UNIX or PC) directory
      in Python's home directory (and name the alias "Dictionary" or "dict", respectively);


USAGE

	>>> from wordnet import *

	>>> # Retrieve words from the database
	>>> N['dog']
	dog(n.)
	>>> V['dog']
	dog(v.)
	>>> ADJ['clear']
	clear(adj.)
	>>> ADV['clearly']
	clearly(adv.)

	>>> # Examine a word's senses and pointers:
	>>> N['dog'].senses()
	('dog' in {noun: dog, domestic dog, Canis familiaris}, 'dog' in {noun: frump, dog}, 'dog' in {noun: dog}, 'dog' in {noun: cad, bounder, blackguard, dog, hound, heel}, 'dog' in {noun: pawl, detent, click, dog}, 'dog' in {noun: andiron, firedog, dog, dogiron})
	>>> # Extract the first sense
	>>> dog = N['dog'][0]	# aka N['dog'].senses()[0]
	>>> dog
	'dog' in {noun: dog, domestic dog, Canis familiaris}
	>>> dog.pointers()
	(hypernym -> {noun: canine, canid}, member meronym -> {noun: Canis, genus Canis}, member meronym -> {noun: pack}, hyponym -> {noun: pooch, doggie, doggy, bow-wow}, hyponym -> {noun: cur, mongrel, mutt}, hyponym -> {noun: lapdog}, hyponym -> {noun: toy dog, toy}, hyponym -> {noun: hunting dog}, hyponym -> {noun: working dog}, hyponym -> {noun: dalmatian, coach dog, carriage dog}, hyponym -> {noun: basenji}, hyponym -> {noun: pug, pug-dog}, hyponym -> {noun: Newfoundland}, hyponym -> {noun: Great Pyrenees}, hyponym -> {noun: spitz}, hyponym -> {noun: griffon, Brussels griffon, Belgian griffon}, hyponym -> {noun: corgi, Welsh corgi}, hyponym -> {noun: poodle, poodle dog}, hyponym -> {noun: Mexican hairless}, part holonym -> {noun: flag})
	>>> dog.pointerTargets(MEMBER_MERONYM)
	[{noun: Canis, genus Canis}, {noun: pack}]


CLASSES

	Dictionary
	Pointer
	Sense
	Synset
	Word


FUNCTIONS

	# database access
	N[string], V[string], ADJ[string], ADV[string]
	for word in N: ...
	getWord(string, pos)

	# cache management
	clearCache()
	disableCache()
	setCacheSize()


GLOBAL VARIABLES

	N: Dictionary
	V: Dictionary
	ADJ: Dictionary
	ADV: Dictionary
	Dictionaries: sequence of all Dictionaries


CONFIGURATION VARIABLES

WNHOME
	Base directory for WordNet. Unix default is /usr/local/wordnet1.6,
	PC default is C:\wn16, Macintosh default is : .
WNSEARCHDIR
	Directory in which the WordNet database has been installed. Unix
	default is WNHOME/dict, PC default is WNHOME\dict , Macintosh default
	is :Database.
ReadableRepresentations
	If true (the default), entities are printed in human-readable form.
	This is necessary because print sequence and str(sequence) request
	the representations of the items, and not their strings.
"""

__author__  = "Oliver Steele", 'steele@cs.brandeis.edu'
__version__ = "1.2"

# Change history:
# 1.0d1	11/02/98
#	initial implementation
# 1.0b1	11/04/98
#	released to RLLC
# 1.0	11/13/98
#	first public release
# 1.1d1	2/27/99
#	simplified copyright notice (it's still free)
#	added cache (buildIndexCaches)
#	make Dictionary print as the variable that's bound to it
#	added method Dictionary.keys(), that returns strings for that part of speech
#	added method Dictionary.has_key()
#	more efficient (don't reopen same pathname)
#	reorganized code: move Cache later, broke _IndexFile out of Dictionary
# 1.1d2	3/8/99
#	added verb frames
# 1.1d3	3/18/99
#	added test code and doctest comments
# 1.1	4/14/99
#	more doctest comments
#	added 'form' attribute as a synonym for 'string'
#	extend closure and its derivatives to work on Words
# 1.1.1	4/21/99
#	fixed len(Dictionary); fixed doctest comments
# 1.1.2	4/24/99
#	fixed for Windows (added workaround for Python file.seek() bug)
# 1.2	4/15/99
#	public release
#	moved nonessential utility functions to wntools
#	renamed private functions to begin with _
#	closure (now in wntools) unions, rather than concatenating
#	fixed searchBinaryFile to find the first line
#
# To do:
# - consolidate Dictionary type and enum (make V and VERB synonyms)
# - document the globals
# - benchmark the index cache; document it or remove it
# - add an override table, to work around the bugs in the 1.6 database files

import string
import os
#from os import environ, path
from types import IntType, ListType, StringType, TupleType


#
# Configuration variables
#
#MTC--for some reason the following original line wasn't working so I'm replacing it with something
#simpler that does work IF you have the environment variables set correctly.
#WNHOME = os.environ.get('WNHOME', {'mac': ":", 'dos': "C:\wn16", 'nt': "C:\wn16"}.get(os.name, "/usr/local/wordnet1.6"))
WNHOME = os.environ['WNHOME']
#WNSEARCHDIR = os.environ.get('WNSEARCHDIR', os.path.join(WNHOME, {'mac': "Database"}.get(os.name, "dict")))
WNSEARCHDIR = os.environ['WNSEARCHDIR']
ReadableRepresentations = 1
_TraceLookups = 0
_FILE_OPEN_MODE = os.name in ('dos', 'nt') and 'rb' or 'r'	# work around a Windows Python bug

#
# Enumerated types
#
NOUN = 'noun'
VERB = 'verb'
ADJECTIVE = 'adjective'
ADVERB = 'adverb'
PartsOfSpeech = (NOUN, VERB, ADJECTIVE, ADVERB)

ANTONYM = 'antonym'
HYPERNYM = 'hypernym'
HYPONYM = 'hyponym'
ATTRIBUTE = 'attribute'
ALSO_SEE = 'also see'
ENTAILMENT = 'entailment'
CAUSE = 'cause'
VERB_GROUP = 'verb group'
MEMBER_MERONYM = 'member meronym'
SUBSTANCE_MERONYM = 'substance meronym'
PART_MERONYM = 'part meronym'
MEMBER_HOLONYM = 'member holonym'
SUBSTANCE_HOLONYM = 'substance holonym'
PART_HOLONYM = 'part holonym'
SIMILAR = 'similar'
PARTICIPLE_OF = 'participle of'
PERTAINYM = 'pertainym'

POINTER_TYPES = (ANTONYM, HYPERNYM, HYPONYM, ATTRIBUTE, ALSO_SEE,
	ENTAILMENT, CAUSE, VERB_GROUP,
	MEMBER_MERONYM, SUBSTANCE_MERONYM, PART_MERONYM,
	MEMBER_HOLONYM, SUBSTANCE_HOLONYM, PART_HOLONYM, SIMILAR,
	PARTICIPLE_OF, PERTAINYM)

ATTRIBUTIVE = 'attributive'
PREDICATIVE = 'predicative'
IMMEDIATE_POSTNOMINAL = 'immediate postnominal'
ADJECTIVE_POSITIONS = (ATTRIBUTIVE, PREDICATIVE, IMMEDIATE_POSTNOMINAL, None)

VERB_FRAME_STRINGS = (
	None,
	"Something %s",
	"Somebody %s",
	"It is %sing",
	"Something is %sing PP",
	"Something %s something Adjective/Noun",
	"Something %s Adjective/Noun",
	"Somebody %s Adjective",
	"Somebody %s something",
	"Somebody %s somebody",
	"Something %s somebody",
	"Something %s something",
	"Something %s to somebody",
	"Somebody %s on something",
	"Somebody %s somebody something",
	"Somebody %s something to somebody",
	"Somebody %s something from somebody",
	"Somebody %s somebody with something",
	"Somebody %s somebody of something",
	"Somebody %s something on somebody",
	"Somebody %s somebody PP",
	"Somebody %s something PP",
	"Somebody %s PP",
	"Somebody's (body part) %s",
	"Somebody %s somebody to INFINITIVE",
	"Somebody %s somebody INFINITIVE",
	"Somebody %s that CLAUSE",
	"Somebody %s to somebody",
	"Somebody %s to INFINITIVE",
	"Somebody %s whether INFINITIVE",
	"Somebody %s somebody into V-ing something",
	"Somebody %s something with something",
	"Somebody %s INFINITIVE",
	"Somebody %s VERB-ing",
	"It %s that CLAUSE",
	"Something %s INFINITIVE")


#
# Domain classes
#
class Word:
	"""An index into the database.

	Each word has one or more Senses, which can be accessed via word.senses()
	or through the index notation, word[n].

	Fields:
		form
			The orthographic representation of the word (a String).
		pos
			The part of speech -- one of NOUN, VERB, ADJECTIVE, ADVERB (a String)
		string
			Same as form (for compatability with version 1.0).
		taggedSenseCount
			The number of senses that are tagged.

	Methods:
		senses()
			Return (sense1, sense2, ...).
		isTagged()
			Return 1 if any sense is tagged.
		adjectivePositions()
			Return a list of adjective labels.
	>>> N['dog'].pos
	'noun'
	>>> N['dog'].string
	'dog'
	>>> N['dog'].taggedSenseCount
	1
	"""

	def __init__(self, line):
		tokens = string.split(line)
		ints = map(int, tokens[int(tokens[3]) + 4:])
		self.string = string.replace(tokens[0], '_', ' ')
		self.form = self.string
		self.pos = _normalizePOS(tokens[1])
		self.taggedSenseCount = ints[1]
		self._synsetOffsets = ints[2:ints[0]+2]

	def senses(self):
		"""Return (sense1, sense2, ...)

		>>> N['dog'].senses()
		('dog' in {noun: dog, domestic dog, Canis familiaris}, 'dog' in {noun: frump, dog}, 'dog' in {noun: dog}, 'dog' in {noun: cad, bounder, blackguard, dog, hound, heel}, 'dog' in {noun: pawl, detent, click, dog}, 'dog' in {noun: andiron, firedog, dog, dogiron})
		"""
		if not hasattr(self, '_senses'):
			def getSense(offset, pos=self.pos, form=self.string):
				return getSynset(pos, offset)[form]
			self._senses = tuple(map(getSense, self._synsetOffsets))
			del self._synsetOffsets
		return self._senses

	def isTagged(self):
		"""Return 1 if any sense is tagged.

		>>> N['dog'].isTagged()
		1
		"""
		return self.taggedSenseCount > 0

	def adjectivePositions(self):
		"""Return a sequence of adjective positions that this word can appear in.

		>>> ADJ['clear'].adjectivePositions()
		['predicative', None]
		"""
		positions = {}
		for sense in self.senses():
			positions[sense.position] = 1
		return positions.keys()

	def __cmp__(self, other):
		"""
		>>> N['cat'] < N['dog']
		1
		>>> N['dog'] < V['dog']
		1
		"""
		return _instanceComparison(self, other, ('pos', 'string'))

	def __str__(self):
		"""Return a human-readable representation.

		>>> str(N['dog'])
		'dog(n.)'
		"""
		abbrs = {NOUN: 'n.', VERB: 'v.', ADJECTIVE: 'adj.', ADVERB: 'adv.'}
		return self.string + "(" + abbrs[self.pos] + ")"

	def __repr__(self):
		"""If ReadableRepresentations is true, return a human-readable
		representation, e.g. 'dog(n.)'.

		If ReadableRepresentations is false, return a machine-readable
		representation, e.g. "getWord('dog', 'noun')".
		"""
		if ReadableRepresentations:
			return str(self)
		return "getWord" + `(self.string, self.pos)`

	#
	# Sequence protocol (a Word's elements are its Senses)
	#
	def __nonzero__(self):
		return 1

	def __len__(self):
		return len(self.senses())

	def __getitem__(self, index):
		return self.senses()[index]

	def __getslice__(self, i, j):
		return self.senses()[i:j]


class Synset:
	"""A set of synonyms that share a common meaning.

	Each synonym contains one or more Senses, which represent a specific sense
	of a specific word.  Senses can be retrieved via synset.senses() or through
	the index notations synset[0], synset[string], or synset[word]. Synsets also
	originate zero or more typed pointers, which can be accessed via
	synset.pointers() or synset.pointers(pointerType).  The targets of a synset
	pointer can be retrieved via synset.pointerTargets() or
	synset.pointerTargets(pointerType), which are equivalent to
	map(Pointer.target, synset.pointerTargets(...)).

	Fields:
		pos
			The part of speech -- one of NOUN, VERB, ADJECTIVE, ADVERB (a String).
		offset
			An integer offset into the part-of-speech file.  Together
			with pos, this can be used as a unique id.
		gloss
			A gloss for the sense (a String).
		verbFrames
			A sequence of integers that index into VERB_FRAME_STRINGS.  These list
			the verb frames that any Sense in this synset participates in.  (See also
			Sense.verbFrames.)  Defined only for verbs.
			>>> V['think'][0].synset.verbFrames
			(5, 9)

	Methods:
		senses()
			Return a sequence of Senses.
		pointers(pointerType=None)
			Return a sequence of Pointers.
		pointerTargets(pointerType=None)
			Return a sequence of Senses or Synsets.
		isTagged()
			Return 1 if any sense is tagged.
	"""

	def __init__(self, pos, offset, line):
		self.pos = pos
		self.offset = offset
		tokens = string.split(line[:string.index(line, '|')])
		self.ssType = tokens[2]
		self.gloss = string.strip(line[string.index(line, '|') + 1:])
		(self._senseTuples, remainder) = _partition(tokens[4:], 2, string.atoi(tokens[3], 16))
		(self._pointerTuples, remainder) = _partition(remainder[1:], 4, int(remainder[0]))
		if pos == VERB:
			(vfTuples, remainder) = _partition(remainder[1:], 3, int(remainder[0]))
			def extractVerbFrames(index, vfTuples):
				return tuple(map(lambda t:string.atoi(t[1]), filter(lambda t,i=index:string.atoi(t[2]) in (0, i), vfTuples)))
			senseVerbFrames = []
			for index in range(1, len(self._senseTuples) + 1):
				senseVerbFrames.append(extractVerbFrames(index, vfTuples))
			self._senseVerbFrames = senseVerbFrames
			self.verbFrames = tuple(extractVerbFrames(None, vfTuples))

	def senses(self):
		"""Return a sequence of Senses.

		>>> N['dog'][0].senses()
		('dog' in {noun: dog, domestic dog, Canis familiaris},)
		"""
		if not hasattr(self, '_senses'):
			def loadSense(senseTuple, verbFrames=None, synset=self):
				return Sense(synset, senseTuple, verbFrames)
			if self.pos == VERB:
				self._senses = tuple(map(loadSense, self._senseTuples, self._senseVerbFrames))
				del self._senseVerbFrames
			else:
				self._senses = tuple(map(loadSense, self._senseTuples))
			del self._senseTuples
		return self._senses

	def pointers(self, pointerType=None):
		"""Return a sequence of Pointers.

		>>> N['dog'][0].pointers()
		(hypernym -> {noun: canine, canid}, member meronym -> {noun: Canis, genus Canis}, member meronym -> {noun: pack}, hyponym -> {noun: pooch, doggie, doggy, bow-wow}, hyponym -> {noun: cur, mongrel, mutt}, hyponym -> {noun: lapdog}, hyponym -> {noun: toy dog, toy}, hyponym -> {noun: hunting dog}, hyponym -> {noun: working dog}, hyponym -> {noun: dalmatian, coach dog, carriage dog}, hyponym -> {noun: basenji}, hyponym -> {noun: pug, pug-dog}, hyponym -> {noun: Newfoundland}, hyponym -> {noun: Great Pyrenees}, hyponym -> {noun: spitz}, hyponym -> {noun: griffon, Brussels griffon, Belgian griffon}, hyponym -> {noun: corgi, Welsh corgi}, hyponym -> {noun: poodle, poodle dog}, hyponym -> {noun: Mexican hairless}, part holonym -> {noun: flag})
		>>> N['dog'][0].pointers(HYPERNYM)
		(hypernym -> {noun: canine, canid},)
		"""
		if not hasattr(self, '_pointers'):
			def loadPointer(tuple, synset=self):
				return Pointer(synset.offset, tuple)
			self._pointers = tuple(map(loadPointer, self._pointerTuples))
			del self._pointerTuples
		if pointerType == None:
			return self._pointers
		else:
			_requirePointerType(pointerType)
			return filter(lambda pointer, type=pointerType: pointer.type == type, self._pointers)

	def pointerTargets(self, pointerType=None):
		"""Return a sequence of Senses or Synsets.

		>>> N['dog'][0].pointerTargets()
		[{noun: canine, canid}, {noun: Canis, genus Canis}, {noun: pack}, {noun: pooch, doggie, doggy, bow-wow}, {noun: cur, mongrel, mutt}, {noun: lapdog}, {noun: toy dog, toy}, {noun: hunting dog}, {noun: working dog}, {noun: dalmatian, coach dog, carriage dog}, {noun: basenji}, {noun: pug, pug-dog}, {noun: Newfoundland}, {noun: Great Pyrenees}, {noun: spitz}, {noun: griffon, Brussels griffon, Belgian griffon}, {noun: corgi, Welsh corgi}, {noun: poodle, poodle dog}, {noun: Mexican hairless}, {noun: flag}]
		>>> N['dog'][0].pointerTargets(HYPERNYM)
		[{noun: canine, canid}]
		"""
		return map(Pointer.target, self.pointers(pointerType))

	def isTagged(self):
		"""Return 1 if any sense is tagged.

		>>> N['dog'][0].isTagged()
		1
		>>> N['dog'][1].isTagged()
		0
		"""
		return len(filter(Sense.isTagged, self.senses())) > 0

	def __str__(self):
		"""Return a human-readable representation.

		>>> str(N['dog'][0].synset)
		'{noun: dog, domestic dog, Canis familiaris}'
		"""
		return "{" + self.pos + ": " + string.joinfields(map(lambda sense:sense.string, self.senses()), ", ") + "}"

	def __repr__(self):
		"""If ReadableRepresentations is true, return a human-readable
		representation, e.g. 'dog(n.)'.

		If ReadableRepresentations is false, return a machine-readable
		representation, e.g. "getSynset(pos, 1234)".
		"""
		if ReadableRepresentations:
			return str(self)
		return "getSynset" + `(self.pos, self.offset)`

	def __cmp__(self, other):
		return _instanceComparison(self, other, ('pos', 'offset'))

	#
	# Sequence protocol (a Synset's elements are its senses).
	#
	def __nonzero__(self):
		return 1

	def __len__(self):
		"""
		>>> len(N['dog'][0].synset)
		3
		"""
		return len(self.senses())

	def __getitem__(self, idx):
		"""
		>>> N['dog'][0].synset[0] == N['dog'][0]
		1
		>>> N['dog'][0].synset['dog'] == N['dog'][0]
		1
		>>> N['dog'][0].synset[N['dog']] == N['dog'][0]
		1
		>>> N['cat'][6]
		'CAT' in {noun: computed tomography, CT, computerized axial tomography, computed axial tomography, CAT}
		"""
		senses = self.senses()
		if isinstance(idx, Word):
			idx = idx.string
		if isinstance(idx, StringType):
			idx = _index(idx, map(lambda sense:sense.string, senses)) or \
				  _index(idx, map(lambda sense:sense.string, senses), _equalsIgnoreCase)
		return senses[idx]

	def __getslice__(self, i, j):
		return self.senses()[i:j]


class Sense:
	"""A specific meaning of a specific word -- the intersection of a Word and a Synset.

	Fields:
		form
			The orthographic representation of the Word this is a Sense of.
		pos
			The part of speech -- one of NOUN, VERB, ADJECTIVE, ADVERB (a String).
		string
			The same as form (for compatability with version 1.0).
		synset
			The Synset that this Sense is a sense of.
		verbFrames
			A sequence of integers that index into VERB_FRAME_STRINGS.  These list
			the verb frames that this Sense partipates in.  Defined only for verbs.
			>>> decide = V['decide'][0].synset	# first synset for 'decide'
			>>> decide[0].verbFrames
			(8, 2, 26, 29)
			>>> decide[1].verbFrames
			(8, 2)
			>>> decide[2].verbFrames
			(8, 26, 29)

	Methods:
		isTagged()
			Return 1 if this sense is tagged.
		pointers(pointerType=None)
			Return a sequence of Pointers.
		pointerTargets(pointerType=None) =>
			Return a sequence of Senses and synsets.
		senses()
			Return (self,).  Provided for consistency with Synset.senses().
		word()
			Return the Word that this is a sense of.
	"""

	def __init__(sense, synset, senseTuple, verbFrames=None):
		# synset is stored by key (pos, synset) rather than object reference, to
		# avoid creating a circular reference between Senses and Synsets that
		# will prevent the vm from garbage-collecting them.
		sense.pos = synset.pos
		sense.synsetOffset = synset.offset
		sense.verbFrames = verbFrames
		(form, idString) = senseTuple
		sense.position = None
		if '(' in form:
			index = string.index(form, '(')
			key = form[index + 1:-1]
			form = form[:index]
			if key == 'a':
				sense.position = ATTRIBUTIVE
			elif key == 'p':
				sense.position = PREDICATIVE
			elif key == 'ip':
				sense.position = IMMEDIATE_POSTNOMINAL
			else:
				raise "unknown attribute " + key
		sense.string = string.replace(form, '_', ' ')
		sense.form = sense.string

	def __getattr__(self, name):
		# see the note at __init__ about why 'synset' is provided as a 'virtual' slot
		if name == 'synset':
			return getSynset(self.pos, self.synsetOffset)
		else:
			raise AttributeError, name

	def __str__(self):
		"""Return a human-readable representation.

		>>> str(N['dog'])
		'dog(n.)'
		"""
		return `self.string` + " in " + str(self.synset)

	def __repr__(self):
		"""If ReadableRepresentations is true, return a human-readable
		representation, e.g. 'dog(n.)'.

		If ReadableRepresentations is false, return a machine-readable
		representation, e.g. "getWord('dog', 'noun')".
		"""
		if ReadableRepresentations:
			return str(self)
		return "%s[%s]" % (`self.synset`, `self.string`)

	def pointers(self, pointerType=None):
		"""Return a sequence of Pointers.

		>>> N['dog'][0].pointers()
		(hypernym -> {noun: canine, canid}, member meronym -> {noun: Canis, genus Canis}, member meronym -> {noun: pack}, hyponym -> {noun: pooch, doggie, doggy, bow-wow}, hyponym -> {noun: cur, mongrel, mutt}, hyponym -> {noun: lapdog}, hyponym -> {noun: toy dog, toy}, hyponym -> {noun: hunting dog}, hyponym -> {noun: working dog}, hyponym -> {noun: dalmatian, coach dog, carriage dog}, hyponym -> {noun: basenji}, hyponym -> {noun: pug, pug-dog}, hyponym -> {noun: Newfoundland}, hyponym -> {noun: Great Pyrenees}, hyponym -> {noun: spitz}, hyponym -> {noun: griffon, Brussels griffon, Belgian griffon}, hyponym -> {noun: corgi, Welsh corgi}, hyponym -> {noun: poodle, poodle dog}, hyponym -> {noun: Mexican hairless}, part holonym -> {noun: flag})
		>>> N['dog'][0].pointers(HYPERNYM)
		(hypernym -> {noun: canine, canid},)
		"""
		senseIndex = _index(self, self.synset.senses())
		def pointsFromThisSense(pointer, selfIndex=senseIndex):
			return pointer.sourceIndex == 0 or pointer.sourceIndex - 1 == selfIndex
		return filter(pointsFromThisSense, self.synset.pointers(pointerType))

	def pointerTargets(self, pointerType=None):
		"""Return a sequence of Senses or Synsets.

		>>> N['dog'][0].pointerTargets()
		[{noun: canine, canid}, {noun: Canis, genus Canis}, {noun: pack}, {noun: pooch, doggie, doggy, bow-wow}, {noun: cur, mongrel, mutt}, {noun: lapdog}, {noun: toy dog, toy}, {noun: hunting dog}, {noun: working dog}, {noun: dalmatian, coach dog, carriage dog}, {noun: basenji}, {noun: pug, pug-dog}, {noun: Newfoundland}, {noun: Great Pyrenees}, {noun: spitz}, {noun: griffon, Brussels griffon, Belgian griffon}, {noun: corgi, Welsh corgi}, {noun: poodle, poodle dog}, {noun: Mexican hairless}, {noun: flag}]
		>>> N['dog'][0].pointerTargets(HYPERNYM)
		[{noun: canine, canid}]
		"""
		return map(Pointer.target, self.pointers(pointerType))

	def senses(self):
		return self,

	def isTagged(self):
		"""Return 1 if any sense is tagged.

		>>> N['dog'][0].isTagged()
		1
		>>> N['dog'][1].isTagged()
		0
		"""
		word = self.word()
		return _index(self, word.senses()) < word.taggedSenseCount

	def word(self):
		return getWord(self.string, self.pos)

	def __cmp__(self, other):
		def senseIndex(sense, synset=self.synset):
			return _index(sense, synset.senses(), testfn=lambda a,b: a.string == b.string)
		return _instanceComparison(self, other, ('synset',)) or cmp(senseIndex(self), senseIndex(other))


class Pointer:
	""" A typed directional relationship between Senses or Synsets.

	Fields:
		type
			One of POINTER_TYPES (a String).
		pos
			The part of speech -- one of NOUN, VERB, ADJECTIVE, ADVERB (a String).

	Methods:
		source()
			Return the sense or synset that this points from.
		target()
			Return the sense or synset that this pointer points to.
	"""

	_POINTER_TYPE_TABLE = {
		'!': ANTONYM, '@': HYPERNYM, '~': HYPONYM,
		'=': ATTRIBUTE, '^': ALSO_SEE, '*': ENTAILMENT, '>': CAUSE,
		'$': VERB_GROUP,
		'#m': MEMBER_MERONYM, '#s': SUBSTANCE_MERONYM, '#p': PART_MERONYM,
		'%m': MEMBER_HOLONYM, '%s': SUBSTANCE_HOLONYM, '%p': PART_HOLONYM,
		'&': SIMILAR, '<': PARTICIPLE_OF, '\\': PERTAINYM}

	def __init__(self, sourceOffset, pointerTuple):
		(type, offset, pos, indices) = pointerTuple
		self.type = Pointer._POINTER_TYPE_TABLE[type]
		self.sourceOffset = sourceOffset
		self.targetOffset = int(offset)
		self.pos = _normalizePOS(pos)
		indices = string.atoi(indices, 16)
		self.sourceIndex = indices >> 8
		self.targetIndex = indices & 255

	def source(self):
		synset = getSynset(self.pos, self.sourceOffset)
		if self.sourceIndex:
			return synset[self.sourceIndex - 1]
		else:
			return synset

	def target(self):
		synset = getSynset(self.pos, self.targetOffset)
		if self.targetIndex:
			return synset[self.targetIndex - 1]
		else:
			return synset

	def __str__(self):
		return self.type + " -> " + str(self.target())

	def __repr__(self):
		if ReadableRepresentations:
			return str(self)
		return "<" + str(self) + ">"

	def __cmp__(self, other):
		diff = _instanceComparison(self, other, ('pos', 'sourceOffset'))
		if diff:
			return diff
		synset = self.source()
		def pointerIndex(sense, synset=synset):
			return _index(sense, synset.pointers(), testfn=lambda a,b: not _instanceComparison(a, b, ('type', 'sourceIndex', 'targetIndex')))
		return cmp(pointerIndex(self), pointerIndex(other))


#
# Dictionary
#
class Dictionary:
	"""A Dictionary contains all the Words in a given part of speech.

	Indexing a dictionary by a string retrieves the word named by that
	string, e.g. dict['dog'].  Indexing by an integer n retrieves the nth word, e.g.
	dict[0].  Access by an arbitrary integer is very slow except in the special
	case where the words are accessed sequentially; this is to support the use
	of dictionaries as the range of a for statement and as the sequence argument
	to map and filter.

	Fields:
		pos
			The part of speech -- one of NOUN, VERB, ADJECTIVE, ADVERB (a String).

	Methods:
		dict.get(String)
			Return the Word whose form is the argument, or None.
		dict[String]
			Return the Word whose form is the argument.  Raises KeyError if no such Word
			exists.
		dict[int n]
			Return the Word indexed by the nth index entry.
			Raises IndexError if the index is invalid.
		has_key(String):
			Return true iff the argument indexes a word in this dictionary.
		keys()
			Return a sorted list of strings that index words in this dictionary.
		len(Dictionary)
			Return the number of index entries (an Integer).
	"""

	def __init__(self, pos, filenameroot):
		self.pos = pos
		self.indexFile = _IndexFile(pos, filenameroot)
		self.dataFile = open(_dataFilePathname(filenameroot), _FILE_OPEN_MODE)

	def __repr__(self):
		dictionaryVariables = {N: 'N', V: 'V', ADJ: 'ADJ', ADV: 'ADV'}
		if dictionaryVariables.get(self):
			return self.__module__ + "." + dictionaryVariables[self]
		return "<%s.%s instance for %s>" % (self.__module__, "Dictionary", self.pos)

	def getWord(self, form, line=None):
		key = string.replace(string.lower(form), ' ', '_')
		pos = self.pos
		def loader(key=key, line=line, indexFile=self.indexFile):
			line = line or indexFile.get(key)
			return line and Word(line)
		word = _entityCache.get((pos, key), loader)
		if word:
			return word
		else:
			raise KeyError, "%s is not in the %s database" % (`form`, `pos`)

	def getSynset(self, offset):
		pos = self.pos
		def loader(pos=pos, offset=offset, dataFile=self.dataFile):
			return Synset(pos, offset, _lineAt(dataFile, offset))
		return _entityCache.get((pos, offset), loader)

	def buildIndexCacheFile(self):
		self.indexFile.buildIndexCacheFile()

	#
	# Sequence protocol (a Dictionary's items are its Words)
	#
	def __nonzero__(self):
		"""Return false.  (This avoids scanning the whole index file to compute len,
		when a Dictionary is used in test position.)

		>>> N and 'true'
		'true'
		"""
		return 1

	def __len__(self):
		"""Return the number of index entries.

		>>> len(ADJ)
		20169
		"""
		if not hasattr(self, 'length'):
			self.length = len(self.indexFile)
		return self.length

	def __getitem__(self, index):
		"""If _index_ is a String, return the Word whose form is _index_.
		If _index_ is an integer _n_, return the Word indexed by the _n_th Word
		in the Index file.

		>>> N['dog']
		dog(n.)
		>>> N[0]
		's gravenhage(n.)
		"""
		if isinstance(index, StringType):
			return self.getWord(index)
		elif isinstance(index, IntType):
			line = self.indexFile[index]
			return self.getWord(string.replace(line[:string.find(line, ' ')], '_', ' '), line)
		else:
			raise TypeError, "%s is not a String or Int" % `index`

	#
	# Dictionary protocol (a Dictionary's values are its words, keyed by their form)
	#
	def get(self, key, default=None):
		"""Return the Word whose form is _key_, or _default_.

		>>> N.get('dog')
		dog(n.)
		>>> N.get('inu')
		"""
		try:
			return self[key]
		except LookupError:
			return default

	def keys(self):
		"""Return a sorted list of strings that index words in this dictionary."""
		return map(lambda key:string.replace(key, ' ', '_'), self.indexFile.keys())

	def has_key(self, form):
		"""Return true iff the argument indexes a word in this dictionary.

		>>> N.has_key('dog')
		1
		>>> N.has_key('inu')
		0
		"""
		return self.indexFile.has_key(form)

	#
	# Testing
	#
	def testKeys(self):
		"""Verify that index lookup can find each word in the index file."""
		print "Testing: ", self
		file = open(self.indexFile.file.name, _FILE_OPEN_MODE)
		counter = 0
		while 1:
			line = file.readline()
			if line == '': break
			if line[0] != ' ':
				key = string.replace(line[:string.find(line, ' ')], '_', ' ')
				if (counter % 1000) == 0:
					print "%s..." % (key,),
					import sys
					sys.stdout.flush()
				counter = counter + 1
				self[key]
		file.close()
		print "done."


class _IndexFile:
	"""An _IndexFile is an implementation class that presents a Sequence and Dictionary
	interface to a sorted index file."""

	def __init__(self, pos, filenameroot):
		self.pos = pos
		self.file = open(_indexFilePathname(filenameroot), _FILE_OPEN_MODE)
		self.offsetLineCache = {}	# Table of (pathname, offset) -> (line, nextOffset)
		self.rewind()
		self.shelfname = os.path.join(WNSEARCHDIR, pos + ".pyidx")
		try:
			import shelve
			self.indexCache = shelve.open(self.shelfname, 'r')
		except:
			pass

	def rewind(self):
		self.file.seek(0)
		while 1:
			offset = self.file.tell()
			line = self.file.readline()
			if (line[0] != ' '):
				break
		self.nextIndex = 0
		self.nextOffset = offset

	#
	# Sequence protocol (an _IndexFile's items are its lines)
	#
	def __nonzero__(self):
		return 1

	def __len__(self):
		if hasattr(self, 'indexCache'):
			return len(self.indexCache)
		self.rewind()
		lines = 0
		while 1:
			line = self.file.readline()
			if line == "":
				break
			lines = lines + 1
		return lines

	def __nonzero__(self):
		return 1

	def __getitem__(self, index):
		if isinstance(index, StringType):
			if hasattr(self, 'indexCache'):
				return self.indexCache[index]
			return binarySearchFile(self.file, index, self.offsetLineCache, 8)
		elif isinstance(index, IntType):
			if hasattr(self, 'indexCache'):
				return self.get(self.keys[index])
			if index < self.nextIndex:
				self.rewind()
			while self.nextIndex <= index:
				self.file.seek(self.nextOffset)
				line = self.file.readline()
				if line == "":
					raise IndexError, "index out of range"
				self.nextIndex = self.nextIndex + 1
				self.nextOffset = self.file.tell()
			return line
		else:
			raise TypeError, "%s is not a String or Int" % `index`

	#
	# Dictionary protocol (an _IndexFile's values are its lines, keyed by the first word)
	#
	def get(self, key, default=None):
		try:
			return self[key]
		except LookupError:
			return default

	def keys(self):
		if hasattr(self, 'indexCache'):
			keys = self.indexCache.keys()
			keys.sort()
			return keys
		else:
			keys = []
			self.rewind()
			while 1:
				line = self.file.readline()
				if not line: break
				keys.append(entryKey(line))
			return keys

	def has_key(self, key):
		if hasattr(self, 'indexCache'):
			return self.indexCache.has_key(key)
		return self.get(key) != None

	#
	# Index file
	#
	def buildIndexCacheFile(self):
		import shelve
		import os
		print "Building %s:" % (self.shelfname,),
		tempname = self.shelfname + ".temp"
		try:
			indexCache = shelve.open(tempname)
			self.rewind()
			count = 0
			while 1:
				offset, line = self.file.tell(), self.file.readline()
				if not line: break
				key = line[:string.find(line, ' ')]
				if (count % 1000) == 0:
					print "%s..." % (key,),
					import sys
					sys.stdout.flush()
				indexCache[key] = line
				count = count + 1
			indexCache.close()
			os.rename(tempname, self.shelfname)
		finally:
			try: os.remove(tempname)
			except: pass
		print "done."
		self.indexCache = shelve.open(self.shelfname, 'r')


#
# Lookup functions
#
def getWord(form, pos='noun'):
	return _dictionaryFor(pos).getWord(form)

def getSense(form, pos='noun', senseno=0):
	return getWord(form, pos)[senseno]

def getSynset(pos, offset):
	return _dictionaryFor(pos).getSynset(offset)

getword = getWord
getsense = getSense
getsynset = getSynset


#
# Private utilities
#
def _requirePointerType(pointerType):
	if pointerType not in POINTER_TYPES:
		raise TypeError, `pointerType` + " is not a pointer type"
	return pointerType

def _instanceComparison(a, b, fields):
	if not hasattr(b, '__class__'):
		return cmp(type(a), type(b))
	elif a.__class__ != b.__class__:
		return cmp(a.__class__, b.__class__)
	for field in fields:
		diff = cmp(getattr(a, field), getattr(b, field))
		if diff:
			return diff
	return 0

def _equalsIgnoreCase(a, b):
	"""Return true iff a and b have the same lowercase representation.

	>>> _equalsIgnoreCase('dog', 'Dog')
	1
	>>> _equalsIgnoreCase('dOg', 'DOG')
	1
	"""
	return a == b or string.lower(a) == string.lower(b)

#
# File utilities
#
def _dataFilePathname(filenameroot):
	if os.name in ('dos', 'nt'):
		return os.path.join(WNSEARCHDIR, filenameroot + ".dat")
	else:
		return os.path.join(WNSEARCHDIR, "data." + filenameroot)

def _indexFilePathname(filenameroot):
	if os.name in ('dos', 'nt'):
		return os.path.join(WNSEARCHDIR, filenameroot + ".idx")
	else:
		return os.path.join(WNSEARCHDIR, "index." + filenameroot)

def binarySearchFile(file, key, cache={}, cacheDepth=-1):
	from stat import ST_SIZE
	key = key + ' '
	keylen = len(key)
	start, end = 0, os.stat(file.name)[ST_SIZE]
	currentDepth = 0
	while start < end:
		middle = (start + end) / 2
		if cache.get(middle):
			(offset, line) = cache[middle]
		else:
			file.seek(max(0, middle - 1))
			if middle > 0:
				file.readline()
			offset, line = file.tell(), file.readline()
			if currentDepth < cacheDepth:
				cache[middle] = (offset, line)
		#print start, middle, end, offset, line,
		if offset > end:
			assert end != middle - 1, "infinite loop"
			end = middle - 1
		elif line[:keylen] == key:# and line[keylen + 1] == ' ':
			return line
		elif line > key:
			assert end != middle - 1, "infinite loop"
			end = middle - 1
		elif line < key:
			start = offset + len(line) - 1
		currentDepth = currentDepth + 1
	return None

def _lineAt(file, offset):
	file.seek(offset)
	return file.readline()


#
# Sequence Utility Functions
#
def _index(key, sequence, testfn=None, keyfn=None):
	"""Return the index of _key_ within _sequence_, using _testfn_ for comparison and
	transforming items of _sequence_ by _keyfn_ first.

	>>> _index('e', 'hello')
	1
	>>> _index('E', 'hello', testfn=_equalsIgnoreCase)
	1
	>>> _index('x', 'hello')
	"""
	index = 0
	for element in sequence:
		value = element
		if keyfn:
			value = keyfn(value)
		if (not testfn and value == key) or (testfn and testfn(value, key)):
			return index
		index = index + 1
	return None

def _partition(sequence, setsize, setcount):
	""" Partition _sequence_ into _setcount_ subsequences of _setsize_ length,
	and a remainder.

	Return (partitions, remainder), where partitions is a sequence of _setcount_
	subsequences of cardinality _setcount_, and apply(append, partitions) +
	remainder == sequence."""

	partitions = []
	for index in range(0, setsize * setcount, setsize):
		partitions.append(sequence[index:index + setsize])
	return (partitions, sequence[setsize * setcount:])


#
# Cache management
#
# Some kind of cache is necessary since Sense -> Synset references are stored by
# key, and it's nice not to have to cons a new copy of a Synset that's been
# paged in each time a Sense's synset is retrieved.  Ideally, we'd use a weak
# dict, but there aren't any.  A strong dict reintroduces the problem that
# eliminating the Sense <-> Synset circularity was intended to resolve:  every
# entity ever seen is preserved forever, making operations that iterate over the
# entire database prohibitive.
#
# The LRUCache approximates a weak dict in the case where temporal locality
# is good.

class _LRUCache:
	""" A cache of values such that least recently used element is flushed when
	the cache fills.

	Methods:
		clear()
		get(key)
		setCapacity(capacity)
	"""

	# private fields:
	# - entities is a dict from key -> (value, timestamp)
	# - history is a dict from timestamp -> key
	# - nextTimeStamp is the timestamp to use with the next value that's added.
	# - oldestTimeStamp is the timestamp of the oldest element (the next one to
	#   remove), or slightly lower than that
	#
	# This lets us retrieve the key given the timestamp, and the timestamp given
	# the key.  (Also the value given either one.)  That's necessary so that we
	# can reorder the history given a key, and also manipulate the values dict
	# given a timestamp.
	#
	# I haven't tried changing history to a List.  An earlier implementation of
	# history as a List was slower than what's here, but the two implementations
	# aren't directly comparable.

	def __init__(this, capacity):
		this.capacity = capacity
		this.clear()

	def clear(this):
		this.values = {}
		this.history = {}
		this.oldestTimestamp = 0
		this.nextTimestamp = 1

	def removeOldestEntry(this):
		while this.oldestTimestamp < this.nextTimestamp:
			if this.history.get(this.oldestTimestamp):
				key = this.history[this.oldestTimestamp]
				del this.history[this.oldestTimestamp]
				del this.values[key]
				return
			this.oldestTimestamp = this.oldestTimestamp + 1

	def setCapacity(this, capacity):
		if capacity == 0:
			this.clear()
		else:
			this.capacity = capacity
			while len(this.values) > this.capacity:
				this.removeOldestEntry()

	def get(this, key, loadfn=None):
		value = None
		if this.values:
			pair = this.values.get(key)
			if pair:
				(value, timestamp) = pair
				del this.history[timestamp]
		if value == None:
			value = loadfn and loadfn()
		if this.values != None:
			timestamp = this.nextTimestamp
			this.nextTimestamp = this.nextTimestamp + 1
			this.values[key] = (value, timestamp)
			this.history[timestamp] = key
			if len(this.values) > this.capacity:
				this.removeOldestEntry()
		return value


class _NullCache:
	"""A NullCache implements the Cache interface (the interface that LRUCache implements),
	but doesn't store any values."""
	def clear():
		pass

	def get(this, key, loadfn=None):
		return loadfn and loadfn()


DEFAULT_CACHE_CAPACITY = 1000
_entityCache = _LRUCache(DEFAULT_CACHE_CAPACITY)

def disableCache():
	"""Disable the entity cache."""
	_entityCache = _NullCache()

def enableCache():
	"""Enable the entity cache."""
	if not isinstance(_entityCache, LRUCache):
		_entityCache = _LRUCache(size)

def clearCache():
	"""Clear the entity cache."""
	_entityCache.clear()

def setCacheCapacity(capacity=DEFAULT_CACHE_CAPACITY):
	"""Set the capacity of the entity cache."""
	enableCache()
	_entityCache.setCapacity(capacity)

setCacheSize = setCacheCapacity	# for compatability with version 1.0


#
# POS Dictionaries (must be initialized after file utilities)
#
N = Dictionary(NOUN, 'noun')
V = Dictionary(VERB, 'verb')
ADJ = Dictionary(ADJECTIVE, 'adj')
ADV = Dictionary(ADVERB, 'adv')
Dictionaries = (N, V, ADJ, ADV)


#
# Part-of-speech tag normalization tables (must be initialized after POS dictionaries)
#
_POSNormalizationTable = {}
_POStoDictionaryTable = {}

def _initializePOSTables():
	global _POSNormalizationTable, _POStoDictionaryTable
	_POSNormalizationTable = {}
	_POStoDictionaryTable = {}
	for pos, abbreviations in (
			(NOUN, "noun n n."),
			(VERB, "verb v v."),
			(ADJECTIVE, "adjective adj adj. a s"),
			(ADVERB, "adverb adv adv. r")):
		tokens = string.split(abbreviations)
		for token in tokens:
			_POSNormalizationTable[token] = pos
			_POSNormalizationTable[string.upper(token)] = pos
	for dict in Dictionaries:
		_POSNormalizationTable[dict] = dict.pos
		_POStoDictionaryTable[dict.pos] = dict

_initializePOSTables()

def _normalizePOS(pos):
	norm = _POSNormalizationTable.get(pos)
	if norm:
		return norm
	raise TypeError, `pos` + " is not a part of speech type"

def _dictionaryFor(pos):
	pos = _normalizePOS(pos)
	dict = _POStoDictionaryTable.get(pos)
	if dict == None:
		raise RuntimeError, "The " + `pos` + " dictionary has not been created"
	return dict

def buildIndexFiles():
	for dict in Dictionaries:
		dict.buildIndexCacheFile()


#
# Testing
#
def _testKeys():
	#This is slow, so don't do it as part of the normal test procedure.
	for dictionary in Dictionaries:
		dictionary.testKeys()

def _test():
	import doctest, wordnet
	# doctest.master = None	# This keeps doctest from complaining after a reload.
	return doctest.testmod(wordnet)
